/**
* <a href="http://servicemix.org">ServiceMix: The open source ESB</a> 
* 
* Copyright 2005 RAJD Consultancy Ltd
* 
* Licensed under the Apache License, Version 2.0 (the "License"); 
* you may not use this file except in compliance with the License. 
* You may obtain a copy of the License at 
* 
* http://www.apache.org/licenses/LICENSE-2.0
* 
* Unless required by applicable law or agreed to in writing, software
* distributed under the License is distributed on an "AS IS" BASIS, 
* WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. 
* See the License for the specific language governing permissions and 
* limitations under the License. 
* 
**/


package org.servicemix.jbi.framework;
import java.util.ArrayList;
import java.util.Collection;
import java.util.HashSet;
import java.util.Iterator;
import java.util.List;
import java.util.Set;
import javax.jbi.JBIException;
import javax.jbi.component.ComponentContext;
import javax.jbi.servicedesc.ServiceEndpoint;
import javax.xml.namespace.QName;
import org.servicemix.jbi.component.ComponentContextImpl;
import org.servicemix.jbi.servicedesc.ServiceEndpointImpl;
import org.w3c.dom.Document;
import org.w3c.dom.DocumentFragment;

/**
 * Registry for Components
 * 
 * @version $Revision: 1.2 $
 */
public class EndpointRegistry {
    private ComponentRegistry componentRegistry;

    /**
     * Constructor
     * 
     * @param cr
     */
    public EndpointRegistry(ComponentRegistry cr) {
        this.componentRegistry = cr;
    }

    /**
     * Get all endpoints for a given service
     * 
     * @param serviceName
     * @return array of endpoints
     */
    public ServiceEndpoint[] getEndpointsForService(QName serviceName) {
        Collection collection = getEndpointsByName(serviceName, getInternalEndpoints());
        return asEndpointArray(collection);
    }

    /**
     * This will return the endpoints for all services and endpoints that implement the named interface (portType in
     * WSDL 1.1). This method does NOT include external endpoints.
     * 
     * @param interfaceName qualified name of interface/portType that is implemented by the endpoint; if
     * <code>null</code> then all activated endpoints in the JBI environment must be returned.
     * @return an array of available endpoints for the specified interface name; must be non-null; may be empty.
     */
    public ServiceEndpoint[] getEndpoints(QName interfaceName) {
        ServiceEndpoint[] result = null;
        List list = new ArrayList();
        for (Iterator i = getInternalEndpoints().iterator();i.hasNext();) {
            ServiceEndpoint endpoint = (ServiceEndpoint) i.next();
            QName[] interfaces = endpoint.getInterfaces();
            if (interfaces != null) {
                if (interfaceName != null) {
                    for (int k = 0;k < interfaces.length;k++) {
                        QName qn = interfaces[k];
                        if (qn != null && qn.equals(interfaceName)) {
                            list.add(endpoint);
                            break;
                        }
                    }
                }
                else {
                    list.add(endpoint);
                }
            }
        }
        result = new ServiceEndpoint[list.size()];
        list.toArray(result);
        return result;
    }

    /**
     * Activate an endpoint
     * 
     * @param provider
     * @param serviceName
     * @param endpointName
     * @return the endpoint
     */
    public ServiceEndpointImpl activateEndpoint(ComponentContextImpl provider, QName serviceName, String endpointName) {
        ServiceEndpointImpl answer = new ServiceEndpointImpl(provider.getComponentNameSpace(), endpointName, serviceName);
        activateEndpoint(provider, answer);
        return answer;
    }

    /**
     * Activate an Endpoint
     * 
     * @param provider
     * @param serviceEndpoint
     */
    public synchronized void activateEndpoint(ComponentContext provider, ServiceEndpointImpl serviceEndpoint) {
        ComponentConnector cc = componentRegistry.getLocalComponentConnector(serviceEndpoint.getComponentNameSpace());
        if (cc != null) {
            cc.addActiveEndpoint(serviceEndpoint);
        }
    }

    /**
     * Called by component context when endpoints are being deactivated.
     * 
     * @param provider
     * @param serviceEndpoint
     */
    public void deactivateEndpoint(ComponentContext provider, ServiceEndpointImpl serviceEndpoint) {
        ComponentConnector cc = componentRegistry.getLocalComponentConnector(serviceEndpoint.getComponentNameSpace());
        if (cc != null) {
            cc.removeActiveEndpoint(serviceEndpoint);
        }
    }

    /**
     * Get the named ServiceEndpoint, if activated
     * 
     * @param service
     * @param name
     * @return the activated ServiceEndpoint or null
     */
    public ServiceEndpointImpl getEndpoint(QName service, String name) {
        ServiceEndpointImpl result = null;
        for (Iterator i = getAllEndpoints().iterator();i.hasNext();) {
            ServiceEndpointImpl endpoint = (ServiceEndpointImpl) i.next();
            if (endpoint.getServiceName().equals(service) && endpoint.getEndpointName().equals(name)) {
                result = endpoint;
                break;
            }
        }
        return result;
    }

    /**
     * Retrieve the service description metadata for the specified endpoint.
     * <p>
     * Note that the result can use either the WSDL 1.1 or WSDL 2.0 description language.
     * 
     * @param endpoint endpoint reference; must be non-null.
     * @return metadata describing endpoint, or <code>null</code> if metadata is unavailable.
     * @exception JBIException invalid endpoint reference.
     */
    public Document getEndpointDescriptor(ServiceEndpoint endpoint) throws JBIException {
        return null;
    }

    /**
     * Registers the given external endpoint with the NMR. This indicates to the NMR that the given endpoint is used as
     * a proxy for external service consumers to access an internal service of the same service name (but a different
     * endpoint name).
     * 
     * @param provider
     * @param externalEndpoint the external endpoint to be registered, must be non-null.
     */
    public void registerExternalEndpoint(ComponentContextImpl provider, ServiceEndpointImpl externalEndpoint) {
        ComponentConnector cc = componentRegistry.getLocalComponentConnector(externalEndpoint.getComponentNameSpace());
        if (cc != null) {
            cc.addExternalActiveEndpoint(externalEndpoint);
        }
    }

    /**
     * Deregisters the given external endpoint with the NMR. This indicates to the NMR that the given external endpoint
     * can no longer be used as a proxy for external service consumers to access an internal service of the same service
     * name.
     * 
     * @param provider
     * @param externalEndpoint the external endpoint to be deregistered; must be non-null.
     */
    public void deregisterExternalEndpoint(ComponentContextImpl provider, ServiceEndpointImpl externalEndpoint) {
        ComponentConnector cc = componentRegistry.getLocalComponentConnector(externalEndpoint.getComponentNameSpace());
        if (cc != null) {
            cc.removeExternalActiveEndpoint(externalEndpoint);
        }
    }

    /**
     * This methods returns only registered external endpoints
     * 
     * @param interfaceName qualified name of interface implemented by the endpoints; must be non-null.
     * @return an array of available external endpoints for the specified interface name; must be non-null; may be
     * empty.
     */
    public ServiceEndpoint[] getExternalEndpoints(QName interfaceName) {
        ServiceEndpoint[] result = null;
        List list = new ArrayList();
        for (Iterator i = getExternalEndpoints().iterator();i.hasNext();) {
            ServiceEndpoint endpoint = (ServiceEndpoint) i.next();
            QName[] interfaces = endpoint.getInterfaces();
            if (interfaces != null) {
                if (interfaceName != null) {
                    for (int k = 0;k < interfaces.length;k++) {
                        QName qn = interfaces[k];
                        if (qn != null && qn.equals(interfaceName)) {
                            list.add(endpoint);
                            break;
                        }
                    }
                }
                else {
                    list.add(endpoint);
                }
            }
        }
        result = new ServiceEndpoint[list.size()];
        list.toArray(result);
        return result;
    }

    /**
     * Get external endpoints for the service
     * 
     * @param serviceName qualified name of service that contains the endpoints; must be non-null.
     * @return an array of available external endpoints for the specified service name; must be non-null; may be empty.
     */
    public ServiceEndpoint[] getExternalEndpointsForService(QName serviceName) {
        ServiceEndpoint[] result = null;
        Set externalEndpoints = getEndpointsByName(serviceName, getExternalEndpoints());
        if (externalEndpoints != null) {
            result = new ServiceEndpoint[externalEndpoints.size()];
            externalEndpoints.toArray(result);
        }
        else {
            result = new ServiceEndpoint[0];
        }
        return result;
    }

    /**
     * Resolve the given endpoint reference into a service endpoint. This is called by the component when it has an EPR
     * that it wants to resolve into a service endpoint.
     * <p>
     * Note that the service endpoint returned refers to a dynamic endpoint; the endpoint will exist only as long as
     * this component retains a strong reference to the object returned by this method. The endpoint may not be included
     * in the list of "activated" endpoints.
     * 
     * @param epr endpoint reference as an XML fragment; must be non-null.
     * @return the service endpoint corresponding to the given endpoint reference; <code>null</code> if the reference
     * cannot be resolved.
     */
    public ServiceEndpointImpl resolveEndpointReference(DocumentFragment epr) {
        return null;
    }

    /**
     * Helper method to convert the given collection into an array of endpoints
     * 
     * @param collection
     * @return array of endpoints
     */
    protected ServiceEndpoint[] asEndpointArray(Collection collection) {
        ServiceEndpoint[] answer = new ServiceEndpoint[collection.size()];
        int count = 0;
        for (Iterator iter = collection.iterator();iter.hasNext();) {
            answer[count++] = (ServiceEndpoint) iter.next();
        }
        return answer;
    }

    /**
     * return a collection of endpoints
     * 
     * @param serviceName
     * @param endpoints
     * @return collection of endpoints
     */
    protected Set getEndpointsByName(QName serviceName, Set endpoints) {
        Set answer = new HashSet();
        for (Iterator i = endpoints.iterator();i.hasNext();) {
            ServiceEndpoint endpoint = (ServiceEndpoint) i.next();
            if (endpoint.getServiceName().equals(serviceName)) {
                answer.add(endpoint);
            }
        }
        return answer;
    }

    /**
     * Utility method to get a ComponentConnector from a serviceName
     * 
     * @param serviceName
     * @return the ComponentConnector
     */
    public ComponentConnector getComponentConnectorByExternalServiceName(QName serviceName) {
        ComponentConnector result = null;
        Set set = getEndpointsByName(serviceName, getAllEndpoints());
        if (!set.isEmpty()) {
            ServiceEndpointImpl endpoint = (ServiceEndpointImpl) set.iterator().next();
            result = componentRegistry.getComponentConnector(endpoint.getComponentNameSpace());
        }
        return result;
    }
    
    public ComponentConnector getComponentConnectorByServiceName(QName serviceName) {
        ComponentConnector result = null;
        Set set = getEndpointsByName(serviceName, getAllEndpoints());
        if (!set.isEmpty()) {
            ServiceEndpointImpl endpoint = (ServiceEndpointImpl) set.iterator().next();
            result = componentRegistry.getComponentConnector(endpoint.getComponentNameSpace());
        }
        return result;
    }

    /**
     * Utility method to get a ComponentConnector from an InterfaceName
     * 
     * @param interfaceName
     * @return the ComponentConnector
     */
    public ComponentConnector getComponentConnector(QName interfaceName) {
        ComponentConnector result = null;
        for (Iterator i = getAllEndpoints().iterator();i.hasNext();) {
            ServiceEndpointImpl endpoint = (ServiceEndpointImpl) i.next();
            QName[] interfaces = endpoint.getInterfaces();
            if (interfaces != null) {
                if (interfaceName != null) {
                    for (int k = 0;k < interfaces.length;k++) {
                        QName qn = interfaces[k];
                        if (qn != null && qn.equals(interfaceName)) {
                            result = componentRegistry.getComponentConnector(endpoint.getComponentNameSpace());
                            break;
                        }
                    }
                }
            }
        }
        return result;
    }

    protected ComponentConnector getComponentConnectorByEndpointName(String endpointName) {
        ComponentConnector result = null;
        if (endpointName != null) {
            for (Iterator i = getAllEndpoints().iterator();i.hasNext();) {
                ServiceEndpointImpl endpoint = (ServiceEndpointImpl) i.next();
                if (endpoint.getEndpointName().equals(endpointName)) {
                    result = componentRegistry.getComponentConnector(endpoint.getComponentNameSpace());
                    break;
                }
            }
        }
        return result;
    }
    
    protected ServiceEndpointImpl getEndpointByName(String endpointName) {
        ServiceEndpointImpl result = null;
        if (endpointName != null) {
            for (Iterator i = getAllEndpoints().iterator();i.hasNext();) {
                ServiceEndpointImpl endpoint = (ServiceEndpointImpl) i.next();
                if (endpoint.getEndpointName().equals(endpointName)) {
                    result = endpoint;
                    break;
                }
            }
        }
        return result;
    }

    /**
     * @return all default endpoints
     */
    protected Set getInternalEndpoints() {
        Set answer = new HashSet();
        for (Iterator iter = this.componentRegistry.getComponentConnectors().iterator();iter.hasNext();) {
            ComponentConnector cc = (ComponentConnector) iter.next();
            answer.addAll(cc.getActiveEndpoints());
        }
        return answer;
    }

    /**
     * @return all external endpoints
     */
    protected Set getExternalEndpoints() {
        Set answer = new HashSet();
        for (Iterator iter = this.componentRegistry.getComponentConnectors().iterator();iter.hasNext();) {
            ComponentConnector cc = (ComponentConnector) iter.next();
            answer.addAll(cc.getExternalActiveEndpoints());
        }
        return answer;
    }

    /**
     * @return all default and external endpoints
     */
    protected Set getAllEndpoints() {
        Set answer = new HashSet();
        for (Iterator iter = this.componentRegistry.getComponentConnectors().iterator();iter.hasNext();) {
            ComponentConnector cc = (ComponentConnector) iter.next();
            answer.addAll(cc.getActiveEndpoints());
            answer.addAll(cc.getExternalActiveEndpoints());
        }
        return answer;
    }
}