/** 
 * 
 * Copyright 2005 LogicBlaze, Inc. http://www.logicblaze.com
 * 
 * Licensed under the Apache License, Version 2.0 (the "License"); 
 * you may not use this file except in compliance with the License. 
 * You may obtain a copy of the License at 
 * 
 * http://www.apache.org/licenses/LICENSE-2.0
 * 
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS, 
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. 
 * See the License for the specific language governing permissions and 
 * limitations under the License. 
 * 
 **/
package org.servicemix;

import junit.framework.TestCase;
import org.servicemix.client.ServiceMixClient;
import org.servicemix.client.ServiceMixClientTest;
import org.servicemix.jbi.container.SpringJBIContainer;
import org.servicemix.jbi.jaxp.SourceTransformer;
import org.servicemix.jbi.util.DOMUtil;
import org.servicemix.jbi.resolver.ServiceNameEndpointResolver;
import org.servicemix.examples.Receiver;
import org.springframework.context.support.AbstractXmlApplicationContext;
import org.springframework.context.support.ClassPathXmlApplicationContext;
import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;
import org.apache.xpath.CachedXPathAPI;
import org.w3c.dom.Node;
import org.w3c.dom.Element;
import org.w3c.dom.traversal.NodeIterator;

import javax.xml.namespace.QName;
import javax.xml.transform.stream.StreamSource;
import javax.xml.transform.TransformerException;
import javax.xml.transform.Source;
import javax.jbi.messaging.MessagingException;
import javax.jbi.messaging.InOnly;
import javax.jbi.messaging.NormalizedMessage;
import javax.jbi.messaging.Fault;
import javax.jbi.JBIException;
import java.io.StringReader;
import java.io.InputStream;
import java.util.List;
import java.util.Iterator;

/**
 * @version $Revision: 1.11 $
 */
public abstract class TestSupport extends TestCase {
    private static final transient Log log = LogFactory.getLog(ServiceMixClientTest.class);
    protected AbstractXmlApplicationContext context;
    protected ServiceMixClient client;
    protected Receiver receiver;
    protected SourceTransformer transformer = new SourceTransformer();
    protected int messageCount = 20;

    protected void setUp() throws Exception {
        context = createBeanFactory();
        context.setXmlValidating(false);

        client = (ServiceMixClient) getBean("client");

        // TODO
        //receiver = (Receiver) getBean("receiver");

        SpringJBIContainer jbi = (SpringJBIContainer) getBean("jbi");
        receiver = (Receiver) jbi.getBean("receiver");
    }

    protected void tearDown() throws Exception {
        if (context != null) {
            System.out.println("Closing down the spring context");
            context.destroy();
        }
    }

    protected Object getBean(String name) {
        Object answer = context.getBean(name);
        assertNotNull("Could not find object in Spring for key: " + name, answer);
        return answer;
    }

    protected abstract AbstractXmlApplicationContext createBeanFactory();

    /**
     * Sends messages to the given service and asserts that the receiver gets them all
     *
     * @param service
     * @throws javax.jbi.messaging.MessagingException
     */
    protected void assertSendAndReceiveMessages(QName service) throws Exception {
        for (int i = 1; i <= messageCount; i++) {
            InOnly exchange = client.createInOnlyExchange();

            NormalizedMessage message = exchange.getInMessage();
            message.setProperty("name", "James");
            message.setContent(new StreamSource(new StringReader("<sample id='" + i + "'>hello world!</sample>")));

            exchange.setService(service);
            client.sendSync(exchange);
            client.done(exchange);

            // lets assert that we have no failure
            Exception error = exchange.getError();
            if (error != null) {
                throw error;
            }

            Fault fault = exchange.getFault();
            assertEquals("Should have no fault!", null, fault);
        }

        assertMessagesReceived();
    }

    protected void assertMessagesReceived() throws TransformerException {
        assertNotNull("receiver not found in JBI container", receiver);

        receiver.getMessageList().assertMessagesReceived(messageCount);
        List list = receiver.getMessageList().getMessages();
        int counter = 0;
        for (Iterator iter = list.iterator(); iter.hasNext();) {
            NormalizedMessage message = (NormalizedMessage) iter.next();
            log.info("Message " + (counter++) + " is: " + message);
            log.info(transformer.toString(message.getContent()));
        }
    }

    /**
     * Performs an XPath expression and returns the Text content of the root node.
     * @param node
     * @param xpath
     * @return
     * @throws TransformerException
     */
    protected String textValueOfXPath(Node node, String xpath) throws TransformerException {
        CachedXPathAPI cachedXPathAPI = new CachedXPathAPI();
        NodeIterator iterator = cachedXPathAPI.selectNodeIterator(node, xpath);
        Node root = iterator.nextNode();
        Element element = (Element) root;
        if (element == null) {
            return "";
        }
        String text = DOMUtil.getElementText(element);
        return text;
    }

    /**
     * Performs a request using the given file from the classpath as the request body and return the answer
     *
     * @param serviceName
     * @param fileOnClassPath
     * @return
     * @throws JBIException
     */
    protected Object requestServiceWithFileRequest(QName serviceName, String fileOnClassPath) throws JBIException {
        Source content = getSourceFromClassPath(fileOnClassPath);

        ServiceNameEndpointResolver resolver = new ServiceNameEndpointResolver(serviceName);

        Object answer = client.request(resolver, null, null, content);
        return answer;
    }

    protected Source getSourceFromClassPath(String fileOnClassPath) {
        InputStream stream = getClass().getResourceAsStream(fileOnClassPath);
        assertNotNull("Could not find file: " + fileOnClassPath + " on the classpath", stream);

        Source content = new StreamSource(stream);
        return content;
    }
}
